import 'dart:async';

import 'package:flutter/material.dart';
import 'package:flutter/services.dart';

final MethodChannel _channel = const MethodChannel('sogou.flutter/videoPlayer')
// This will clear all open videos on the platform when a full restart is
// performed.
  ..invokeMethod<void>('init');

class DurationRange {
  DurationRange(this.start, this.end);

  final Duration start;
  final Duration end;

  double startFraction(Duration duration) {
    return start.inMilliseconds / duration.inMilliseconds;
  }

  double endFraction(Duration duration) {
    return end.inMilliseconds / duration.inMilliseconds;
  }

  @override
  String toString() => '$runtimeType(start: $start, end: $end)';
}

enum PlayerStatus {
  unknown, // VideoPlayerController instance was created; no native method was called
  created, // native player created
  initialized, // video info can be fetched: size, during etc.
  playing,
  paused,
  completed,
  error, // including plugin error and native player error when playing
}

class VideoPlayerEvent {
  final PlayerStatus status;
  final bool pending; // 是否在等待加载
  final Duration duration;
  final Size size;

  VideoPlayerEvent(this.status, this.pending, this.duration, this.size);

  VideoPlayerEvent copyWith({
    PlayerStatus status,
    bool pending,
    Duration duration,
    Size size,
  }) {
    return VideoPlayerEvent(
      status ?? this.status,
      pending ?? this.pending,
      duration ?? this.duration,
      size ?? this.size,
    );
  }

  @override
  String toString() => 'VideoPlayerEvent($status, $pending, $duration, $size)';
}

class VideoPlayerController extends ValueNotifier<VideoPlayerEvent> {
  int _textureId;
  String dataSource;

  String package;
  bool _isDisposed = false;
  StreamSubscription<dynamic> _eventSubscription;

  VideoPlayerController()
      : super(VideoPlayerEvent(
      PlayerStatus.unknown, true, Duration.zero, Size.zero));

  Duration get pausedDuration {
    return Duration.zero;
  }

  // init should be invoked at first time;
  Future<void> init() async {
    if (_textureId != null) {
      print("No need to re-init for videocontroller:$_textureId");
      return;
    }
    // 主要是生成textureId，以供后续使用
    Map<String, dynamic> response =
    await _channel.invokeMapMethod<String, dynamic>('create');

    // 再做一次检查，避免重入
    if (_textureId != null) {
      print(
          "init was called twice, Use first texureID instead:$_textureId");
      return;
    }

    _textureId = response['textureId'];
    // 如果 native各种原因，创建失败，走PlatformException
    assert(_textureId != null);

    _eventSubscription = _eventChannelFor(_textureId)
        .receiveBroadcastStream()
        .listen(_eventListener, onError: _errorListener);

    // 更新状态
    value = value.copyWith(status: PlayerStatus.created);
  }

  void _errorListener(Object obj) {
    print("errorListener:$obj");
    final PlatformException e = obj;
    value = value.copyWith(status: PlayerStatus.error);
  }

  void _eventListener(dynamic event) {
    final Map<dynamic, dynamic> map = event;
    print("event-$_textureId=$map");
    String eventType = map['event'];
    switch (eventType) {
      case 'initialized':
        _handleInitEvent(map);
        break;
      case 'playing':
        value = value.copyWith(status: PlayerStatus.playing, pending: false);
        break;
      case 'paused':
        value = value.copyWith(status: PlayerStatus.paused, pending: false);
        break;
      case 'completed':
        value = value.copyWith(status: PlayerStatus.completed, pending: false);
        break;
      case 'pending':
        bool isPending = map['isPending'];
        assert(isPending != null);
        value = value.copyWith(pending: isPending);
        break;
      case 'error':
        value = value.copyWith(status: PlayerStatus.error);
        break;
      case 'created':
      // 应用层在调用了 stop 函数后，native应该返回此状态；注：亦有可能是 error
        value = value.copyWith(
            status: PlayerStatus.created,
            pending: false,
            duration: Duration.zero,
            size: Size.zero);
        break;
      default:
        print("Unknown player event type:$eventType");
        break;
    }
  }

  Future<void> resetDataSource(String newDataSource,
      [bool autoPlay = false]) async {
    dataSource = newDataSource;
    // 更新状态
    value = value.copyWith(status: PlayerStatus.created);
    // native，需要及时返回新的状态，如果有错误，native要清理错误
    await _channel.invokeMethod<void>(
      'reset',
      <String, dynamic>{
        'textureId': _textureId,
        'uri': newDataSource,
        'autoPlay': autoPlay
      },
    );
  }

  void _handleInitEvent(Map<dynamic, dynamic> map) {
    int durationLong = map['duration'];
    int width = map['width'];
    int height = map['height'];
    assert(durationLong != null && durationLong >= 0);
    assert(width != null && width >= 0);
    assert(height != null && height >= 0);

    value = value.copyWith(
      status: PlayerStatus.initialized,
      duration: Duration(milliseconds: durationLong),
      size: Size(width?.toDouble(), height?.toDouble()),
      pending: false,
    );
  }

  EventChannel _eventChannelFor(int textureId) {
    return EventChannel('sogou.flutter/videoPlayer/videoEvents$textureId');
  }

  @override
  Future<void> dispose() async {
    if (_textureId != null && !_isDisposed) {
      _isDisposed = true;
      await _eventSubscription?.cancel();
      await _channel.invokeMethod<void>(
        'dispose',
        <String, dynamic>{'textureId': _textureId},
      );
      _textureId = null;
    }
    _isDisposed = true;
    super.dispose();
  }

  // rewind 标示是否从开头播放
  Future<void> play([rewind = false]) async {
    if (_isDisposed ||
        value.status == PlayerStatus.unknown ||
        value.status == PlayerStatus.created) {
      print("No need to sync paly status; not work for 'created' & 'unknown' ");
      return;
    }

    assert(_textureId != null);

    // native，需要及时返回新的状态
    await _channel.invokeMethod<void>(
      'play',
      <String, dynamic>{'textureId': _textureId, 'rewind': rewind},
    );
//    value = value.copyWith(pending: false);
  }

  Future<void> setLooping(bool looping) async {
    // native需要保存looping状态；因为有可能尚未创建成功，应用层就设置looping；等创建成功后，native需要复用looping状态
    if (_isDisposed || value.status == PlayerStatus.unknown) {
      print("No need to sync looping status:$looping ,not work for 'unknown' ");
      return;
    }

    assert(_textureId != null);
    await _channel.invokeMethod<void>(
      'setLooping',
      <String, dynamic>{'textureId': _textureId, 'looping': looping},
    );
  }

  Future<void> pause() async {
    if (_isDisposed ||
        value.status == PlayerStatus.unknown ||
        value.status == PlayerStatus.created ||
        value.status == PlayerStatus.error) {
      print(
          "No need to sync pause status; not work for 'unknown' ,'created' ,'error' status");
      return;
    }

    assert(_textureId != null);
    // native，需要及时返回新的状态
    await _channel.invokeMethod<void>(
      'pause',
      <String, dynamic>{'textureId': _textureId},
    );
  }

  Future<void> stop() async {
    if (_isDisposed ||
        value.status == PlayerStatus.unknown ||
        value.status == PlayerStatus.created) {
      print(
          "No need to sync stop status; not work for 'unknown' ,'created' status");
      return;
    }

    assert(_textureId != null);
    // native，需要及时返回新的状态
    await _channel.invokeMethod<void>(
      'stop',
      <String, dynamic>{'textureId': _textureId},
    );
  }

  Future<void> seekTo(Duration moment) async {
    if (_isDisposed ||
        value.status == PlayerStatus.unknown ||
        value.status == PlayerStatus.created ||
        value.status == PlayerStatus.error) {
      print("Can not sync seekTo operaion for 'unknown','created' and 'error'");
      return;
    }

    if (moment > value.duration) {
      moment = value.duration;
    } else if (moment < const Duration()) {
      moment = const Duration();
    }
    assert(_textureId != null);
    // seek之后，native有可能重新启动缓存；缓存结束收，native会通过event通知变更pending状态；
    // 此处，直接将pending状态设置为true
//    value = value.copyWith(pending: true);

    await _channel.invokeMethod<void>('seekTo', <String, dynamic>{
      'textureId': _textureId,
      'location': moment.inMilliseconds,
    });
  }

  /// Sets the audio volume of [this].
  ///
  /// [volume] indicates a value between 0.0 (silent) and 1.0 (full volume) on a
  /// linear scale.
  Future<void> setVolume(double volume) async {
    if (_isDisposed ||
        value.status == PlayerStatus.unknown ||
        value.status == PlayerStatus.created ||
        value.status == PlayerStatus.error) {
      print("setVolume does not works for 'unknown', 'created' and 'error'");
      return;
    }
    assert(_textureId != null);
    await _channel.invokeMethod<void>(
      'setVolume',
      <String, dynamic>{'textureId': _textureId, 'volume': volume},
    );
  }

  // 获取native错误信息
  Future<String> errorMsg() async {
    return await _channel.invokeMethod<String>(
      'errorMsg',
      <String, dynamic>{'textureId': _textureId},
    );
  }

  /// The position in the current video.
  Future<Duration> get position async {
    if (_isDisposed) {
      return null;
    }
    assert(_textureId != null);
    return Duration(
      milliseconds: await _channel.invokeMethod<int>(
        'position',
        <String, dynamic>{'textureId': _textureId},
      ),
    );
  }
}

/// Displays the video controlled by [controller].
class VideoPlayer extends StatefulWidget {
  VideoPlayer(this.controller);

  final VideoPlayerController controller;

  @override
  _VideoPlayerState createState() => _VideoPlayerState();
}

class _VideoPlayerState extends State<VideoPlayer> {
  _VideoPlayerState() {
    _listener = () {
      final int newTextureId = widget.controller._textureId;
      if (newTextureId != _textureIdOld) {
        if (!mounted) {
          return;
        }
        setState(() {
          _textureIdOld = newTextureId;
        });
      }
    };
  }

  VoidCallback _listener;
  int _textureIdOld;

  @override
  void initState() {
    super.initState();
    _textureIdOld = widget.controller._textureId;
    widget.controller.addListener(_listener);
  }

  @override
  void dispose() {
    widget.controller.removeListener(_listener);
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return _textureIdOld == null
        ? Container()
        : Texture(textureId: _textureIdOld);
  }
}
